import asyncio
import logging
from typing import Any, Generic, Protocol, TypeVar

from .agent_message import AgentMessage
from .run_context import CtxT, RunContextWrapper
from .typing.io import AgentID, AgentState

logger = logging.getLogger(__name__)


_MH_PayloadT = TypeVar("_MH_PayloadT", contravariant=True)  # noqa: PLC0105
_MH_StateT = TypeVar("_MH_StateT", bound=AgentState, contravariant=True)  # noqa: PLC0105


class MessageHandler(Protocol[_MH_PayloadT, _MH_StateT, CtxT]):
    async def __call__(
        self,
        message: AgentMessage[_MH_PayloadT, _MH_StateT],
        ctx: RunContextWrapper[CtxT] | None,
        **kwargs: Any,
    ) -> None: ...


class AgentMessagePool(Generic[CtxT]):
    def __init__(self) -> None:
        self._queues: dict[AgentID, asyncio.Queue[AgentMessage[Any, AgentState]]] = {}
        self._message_handlers: dict[
            AgentID, MessageHandler[Any, AgentState, CtxT]
        ] = {}
        self._tasks: dict[AgentID, asyncio.Task[None]] = {}

    async def post(self, message: AgentMessage[Any, AgentState]) -> None:
        for recipient_id in message.recipient_ids:
            queue = self._queues.setdefault(recipient_id, asyncio.Queue())
            await queue.put(message)

    def register_message_handler(
        self,
        agent_id: AgentID,
        handler: MessageHandler[Any, AgentState, CtxT],
        ctx: RunContextWrapper[CtxT] | None = None,
        **run_kwargs: Any,
    ) -> None:
        self._message_handlers[agent_id] = handler
        self._queues.setdefault(agent_id, asyncio.Queue())
        if agent_id not in self._tasks:
            self._tasks[agent_id] = asyncio.create_task(
                self._process_messages(agent_id, ctx=ctx, **run_kwargs)
            )

    async def _process_messages(
        self,
        agent_id: AgentID,
        ctx: RunContextWrapper[CtxT] | None = None,
        **run_kwargs: Any,
    ) -> None:
        queue = self._queues[agent_id]
        while True:
            try:
                message = await queue.get()
                handler = self._message_handlers.get(agent_id)
                if handler is None:
                    break

                try:
                    await self._message_handlers[agent_id](
                        message, ctx=ctx, **run_kwargs
                    )
                except Exception:
                    logger.exception(f"Error handling message for {agent_id}")

                queue.task_done()

            except Exception:
                logger.exception(f"Unexpected error in processing loop for {agent_id}")

    async def unregister_message_handler(self, agent_id: AgentID) -> None:
        if task := self._tasks.get(agent_id):
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.debug(f"{agent_id} exited")

        self._tasks.pop(agent_id, None)
        self._queues.pop(agent_id, None)
        self._message_handlers.pop(agent_id, None)

    async def stop_all(self) -> None:
        for agent_id in list(self._tasks):
            await self.unregister_message_handler(agent_id)



from collections.abc import Sequence
from logging import getLogger
from typing import Any, ClassVar, Generic, Protocol, TypeVar, cast, final

from ..agent_message_pool import AgentMessage, AgentMessagePool
from ..comm_agent import CommunicatingAgent
from ..run_context import CtxT, RunContextWrapper
from ..typing.io import AgentID, AgentState, InT, OutT
from .workflow_agent import WorkflowAgent

logger = getLogger(__name__)

_EH_OutT = TypeVar("_EH_OutT", contravariant=True)  # noqa: PLC0105


class ExitWorkflowLoopHandler(Protocol[_EH_OutT, CtxT]):
    def __call__(
        self,
        output_message: AgentMessage[_EH_OutT, Any],
        ctx: RunContextWrapper[CtxT] | None,
        **kwargs: Any,
    ) -> bool: ...


class LoopedWorkflowAgent(WorkflowAgent[InT, OutT, CtxT], Generic[InT, OutT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        agent_id: AgentID,
        subagents: Sequence[CommunicatingAgent[Any, Any, Any, CtxT]],
        exit_agent: CommunicatingAgent[Any, OutT, Any, CtxT],
        message_pool: AgentMessagePool[CtxT] | None = None,
        recipient_ids: list[AgentID] | None = None,
        dynamic_routing: bool = False,
        max_iterations: int = 10,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        super().__init__(
            subagents=subagents,
            agent_id=agent_id,
            start_agent=subagents[0],
            end_agent=exit_agent,
            message_pool=message_pool,
            recipient_ids=recipient_ids,
            dynamic_routing=dynamic_routing,
        )

        self._max_iterations = max_iterations

        self._exit_workflow_loop_impl: ExitWorkflowLoopHandler[OutT, CtxT] | None = None

    @property
    def max_iterations(self) -> int:
        return self._max_iterations

    def exit_workflow_loop_handler(
        self, func: ExitWorkflowLoopHandler[OutT, CtxT]
    ) -> ExitWorkflowLoopHandler[OutT, CtxT]:
        self._exit_workflow_loop_impl = func

        return func

    def _exit_workflow_loop(
        self,
        output_message: AgentMessage[OutT, Any],
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        **kwargs: Any,
    ) -> bool:
        if self._exit_workflow_loop_impl:
            return self._exit_workflow_loop_impl(output_message, ctx=ctx, **kwargs)

        return False

    @final
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        rcv_args: InT | Sequence[InT] | None = None,
        rcv_message: AgentMessage[InT, Any] | None = None,
        ctx: RunContextWrapper[CtxT] | None = None,
        entry_point: bool = False,
        forbid_state_change: bool = False,
        **kwargs: Any,
    ) -> AgentMessage[OutT, AgentState]:
        agent_message = rcv_message
        num_iterations = 0
        exit_message: AgentMessage[OutT, Any] | None = None

        while True:
            for subagent in self.subagents:
                agent_message = await subagent.run(
                    chat_inputs=chat_inputs,
                    rcv_args=rcv_args,
                    rcv_message=agent_message,
                    entry_point=entry_point,
                    forbid_state_change=forbid_state_change,
                    ctx=ctx,
                    **kwargs,
                )

                if subagent is self._end_agent:
                    num_iterations += 1
                    exit_message = cast("AgentMessage[OutT, AgentState]", agent_message)
                    if self._exit_workflow_loop(exit_message, ctx=ctx):
                        return exit_message
                    if num_iterations >= self._max_iterations:
                        logger.info(
                            f"Max iterations reached ({self._max_iterations}). Exiting loop."
                        )
                        return exit_message

                chat_inputs = None
                rcv_args = None
                entry_point = False



import inspect
from collections.abc import Callable, Coroutine, Sequence
from typing import Any, overload

from .types import P, ProcessorCallableList, ProcessorCallableSingle, R, T


def is_bound_method(func: Callable[..., Any], self_candidate: Any) -> bool:
    return (inspect.ismethod(func) and (func.__self__ is self_candidate)) or hasattr(
        self_candidate, func.__name__
    )


@overload
def split_pos_args(
    call: ProcessorCallableSingle[T, P, R],
    args: Sequence[Any],
) -> tuple[Any | None, T, Sequence[Any]]: ...


@overload
def split_pos_args(
    call: ProcessorCallableList[T, P, R],
    args: Sequence[Any],
) -> tuple[Any | None, list[T], Sequence[Any]]: ...


def split_pos_args(
    call: (ProcessorCallableSingle[T, P, R] | ProcessorCallableList[T, P, R]),
    args: Sequence[Any],
) -> tuple[Any | None, T | list[T], Sequence[Any]]:
    if not args:
        raise ValueError("No positional arguments passed.")
    maybe_self = args[0]
    if is_bound_method(call, maybe_self):
        # Case: Bound instance method with signature (self, inp, *rest)
        if len(args) < 2:
            raise ValueError(
                "Must pass at least `self` and an input (or a list of inputs) "
                "for a bound instance method."
            )
        return maybe_self, args[1], args[2:]
    # Case: Standalone function with signature (inp, *rest)
    if not args:
        raise ValueError(
            "Must pass an input (or a list of inputs) " + "for a standalone function."
        )
    return None, args[0], args[1:]


def partial_processor_callable(
    call: Callable[..., Coroutine[Any, Any, R]],
    self_obj: Any,
    *args: Any,
    **kwargs: Any,
) -> Callable[[Any], Coroutine[Any, Any, R]]:
    async def wrapper(inp: Any) -> R:
        if self_obj is not None:
            # `call` is a method
            return await call(self_obj, inp, *args, **kwargs)
        # `call` is a function
        return await call(inp, *args, **kwargs)

    return wrapper



import httpx
from pydantic import BaseModel, NonNegativeFloat, PositiveInt


class AsyncHTTPClientParams(BaseModel):
    timeout: NonNegativeFloat = 10
    max_connections: PositiveInt = 2000
    max_keepalive_connections: PositiveInt = 500
    keepalive_expiry: float | None = 5


def create_async_http_client(
    client_params: AsyncHTTPClientParams,
) -> httpx.AsyncClient:
    http_client = httpx.AsyncClient(
        timeout=httpx.Timeout(client_params.timeout),
        limits=httpx.Limits(
            max_connections=client_params.max_connections,
            max_keepalive_connections=client_params.max_keepalive_connections,
            keepalive_expiry=client_params.keepalive_expiry,
        ),
    )

    return http_client



from collections.abc import Iterable

from ..typing.content import (
    Content,
    ContentPart,
    ContentPartImage,
    ContentPartText,
    ImageData,
)
from . import (
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionImageURL,
)

BASE64_PREFIX = "data:image/jpeg;base64,"


def image_data_to_str(image_data: ImageData) -> str:
    if image_data.type == "url":
        return str(image_data.url)
    if image_data.type == "base64":
        return f"{BASE64_PREFIX}{image_data.base64}"
    raise ValueError(f"Unsupported image data type: {image_data.type}")


def from_api_content(
    api_content: str | Iterable[ChatCompletionContentPartParam],
) -> "Content":
    if isinstance(api_content, str):
        return Content(parts=[ContentPartText(data=api_content)])

    content_parts: list[ContentPart] = []
    for api_content_part in api_content:
        content_part: ContentPart

        if api_content_part["type"] == "text":
            text_data = api_content_part["text"]
            content_part = ContentPartText(data=text_data)

        elif api_content_part["type"] == "image_url":
            url = api_content_part["image_url"]["url"]
            detail = api_content_part["image_url"].get("detail")
            if url.startswith(BASE64_PREFIX):
                image_data = ImageData.from_base64(
                    base64_encoding=url.removeprefix(BASE64_PREFIX),
                    detail=detail,
                )
            else:
                image_data = ImageData.from_url(img_url=url, detail=detail)  # type: ignore
            content_part = ContentPartImage(data=image_data)

        content_parts.append(content_part)  # type: ignore

    return Content(parts=content_parts)


def to_api_content(
    content: Content,
) -> Iterable[ChatCompletionContentPartParam]:
    api_content: list[ChatCompletionContentPartParam] = []
    for content_part in content.parts:
        api_content_part: ChatCompletionContentPartParam
        if isinstance(content_part, ContentPartText):
            api_content_part = ChatCompletionContentPartTextParam(
                type="text",
                text=content_part.data,
            )
        else:
            api_content_part = ChatCompletionContentPartImageParam(
                type="image_url",
                image_url=ChatCompletionImageURL(
                    url=image_data_to_str(content_part.data),
                    detail=content_part.data.detail,
                ),
            )
        api_content.append(api_content_part)

    return api_content



from collections.abc import Sequence
from typing import Generic, TypeVar
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field

from .typing.io import AgentID, AgentState

_PayloadT = TypeVar("_PayloadT", covariant=True)  # noqa: PLC0105
_StateT = TypeVar("_StateT", bound=AgentState, covariant=True)  # noqa: PLC0105


class AgentMessage(BaseModel, Generic[_PayloadT, _StateT]):
    payloads: Sequence[_PayloadT]
    sender_id: AgentID
    sender_state: _StateT | None = None
    recipient_ids: Sequence[AgentID] = Field(default_factory=list)

    message_id: str = Field(default_factory=lambda: str(uuid4())[:8])

    model_config = ConfigDict(extra="forbid", frozen=True)

    def __repr__(self) -> str:
        return (
            f"From: {self.sender_id}, To: {', '.join(self.recipient_ids)}, "
            f"Payloads: {len(self.payloads)}"
        )



from collections.abc import Sequence
from typing import Any, ClassVar, Generic, cast, final

from ..agent_message_pool import AgentMessage, AgentMessagePool
from ..comm_agent import CommunicatingAgent
from ..run_context import CtxT, RunContextWrapper
from ..typing.io import AgentID, InT, OutT
from .workflow_agent import WorkflowAgent


class SequentialWorkflowAgent(WorkflowAgent[InT, OutT, CtxT], Generic[InT, OutT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        agent_id: AgentID,
        subagents: Sequence[CommunicatingAgent[Any, Any, Any, CtxT]],
        message_pool: AgentMessagePool[CtxT] | None = None,
        recipient_ids: list[AgentID] | None = None,
        dynamic_routing: bool = False,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        super().__init__(
            subagents=subagents,
            start_agent=subagents[0],
            end_agent=subagents[-1],
            agent_id=agent_id,
            message_pool=message_pool,
            recipient_ids=recipient_ids,
            dynamic_routing=dynamic_routing,
        )

    @final
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        rcv_args: InT | Sequence[InT] | None = None,
        rcv_message: AgentMessage[InT, Any] | None = None,
        ctx: RunContextWrapper[CtxT] | None = None,
        entry_point: bool = False,
        forbid_state_change: bool = False,
        **kwargs: Any,
    ) -> AgentMessage[OutT, Any]:
        agent_message = rcv_message
        for subagent in self.subagents:
            agent_message = await subagent.run(
                chat_inputs=chat_inputs,
                rcv_args=rcv_args,
                rcv_message=agent_message,
                entry_point=entry_point,
                forbid_state_change=forbid_state_change,
                ctx=ctx,
                **kwargs,
            )
            chat_inputs = None
            rcv_args = None
            entry_point = False

        return cast("AgentMessage[OutT, Any]", agent_message)



import logging
from collections.abc import Iterable
from copy import deepcopy
from typing import Any, Literal

from openai import AsyncOpenAI
from openai._types import NOT_GIVEN  # type: ignore[import]
from pydantic import BaseModel

from ..cloud_llm import CloudLLM, CloudLLMSettings
from ..http_client import AsyncHTTPClientParams
from ..rate_limiting.rate_limiter_chunked import RateLimiterC
from ..typing.message import AssistantMessage, Conversation
from ..typing.tool import BaseTool
from . import (
    ChatCompletion,
    ChatCompletionAsyncStream,  # type: ignore[import]
    ChatCompletionChunk,
    ChatCompletionMessageParam,
    ChatCompletionPredictionContentParam,
    ChatCompletionStreamOptionsParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
    ParsedChatCompletion,
    # ResponseFormatJSONObject,
    # ResponseFormatJSONSchema,
    # ResponseFormatText,
)
from .converters import OpenAIConverters

logger = logging.getLogger(__name__)


class OpenAILLMSettings(CloudLLMSettings, total=False):
    reasoning_effort: Literal["low", "medium", "high"] | None

    parallel_tool_calls: bool

    # response_format: (
    #     ResponseFormatText | ResponseFormatJSONSchema | ResponseFormatJSONObject
    # )

    modalities: list[Literal["text", "audio"]] | None

    frequency_penalty: float | None
    presence_penalty: float | None
    logit_bias: dict[str, int] | None
    stop: str | list[str] | None
    logprobs: bool | None
    top_logprobs: int | None
    n: int | None

    prediction: ChatCompletionPredictionContentParam | None

    stream_options: ChatCompletionStreamOptionsParam | None

    metadata: dict[str, str] | None
    store: bool | None
    user: str


class OpenAILLM(CloudLLM[OpenAILLMSettings, OpenAIConverters]):
    def __init__(
        self,
        # Base LLM args
        model_name: str,
        model_id: str | None = None,
        llm_settings: OpenAILLMSettings | None = None,
        tools: list[BaseTool[BaseModel, Any, Any]] | None = None,
        response_format: type | None = None,
        # Connection settings
        async_http_client_params: (
            dict[str, Any] | AsyncHTTPClientParams | None
        ) = None,
        async_openai_client_params: dict[str, Any] | None = None,
        # Rate limiting
        rate_limiter: (RateLimiterC[Conversation, AssistantMessage] | None) = None,
        rate_limiter_rpm: float | None = None,
        rate_limiter_chunk_size: int = 1000,
        rate_limiter_max_concurrency: int = 300,
        # Retries
        num_generation_retries: int = 0,
        # Disable tqdm for batch processing
        no_tqdm: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            model_name=model_name,
            model_id=model_id,
            llm_settings=llm_settings,
            converters=OpenAIConverters(),
            tools=tools,
            response_format=response_format,
            async_http_client_params=async_http_client_params,
            rate_limiter=rate_limiter,
            rate_limiter_rpm=rate_limiter_rpm,
            rate_limiter_chunk_size=rate_limiter_chunk_size,
            rate_limiter_max_concurrency=rate_limiter_max_concurrency,
            num_generation_retries=num_generation_retries,
            no_tqdm=no_tqdm,
            **kwargs,
        )

        async_openai_client_params_ = deepcopy(async_openai_client_params or {})
        if self._async_http_client is not None:
            async_openai_client_params_["http_client"] = self._async_http_client

        # TODO: context manager for async client
        self._client: AsyncOpenAI = AsyncOpenAI(
            base_url=self._base_url,
            api_key=self._api_key,
            **async_openai_client_params_,
        )

    async def _get_completion(
        self,
        api_messages: Iterable[ChatCompletionMessageParam],
        api_tools: list[ChatCompletionToolParam] | None = None,
        api_tool_choice: ChatCompletionToolChoiceOptionParam | None = None,
        **api_llm_settings: Any,
    ) -> ChatCompletion:
        tools = api_tools or NOT_GIVEN
        tool_choice = api_tool_choice or NOT_GIVEN

        return await self._client.chat.completions.create(
            model=self._api_model_name,
            messages=api_messages,
            tools=tools,
            tool_choice=tool_choice,
            stream=False,
            **api_llm_settings,
        )

    async def _get_parsed_completion(
        self,
        api_messages: Iterable[ChatCompletionMessageParam],
        api_tools: list[ChatCompletionToolParam] | None = None,
        api_tool_choice: ChatCompletionToolChoiceOptionParam | None = None,
        api_response_format: type | None = None,
        **api_llm_settings: Any,
    ) -> ParsedChatCompletion[Any]:
        tools = api_tools or NOT_GIVEN
        tool_choice = api_tool_choice or NOT_GIVEN
        response_format = api_response_format or NOT_GIVEN

        return await self._client.beta.chat.completions.parse(
            model=self._api_model_name,
            messages=api_messages,
            tools=tools,
            tool_choice=tool_choice,
            response_format=response_format,  # type: ignore[arg-type]
            **api_llm_settings,
        )

    async def _get_completion_stream(
        self,
        api_messages: Iterable[ChatCompletionMessageParam],
        api_tools: list[ChatCompletionToolParam] | None = None,
        api_tool_choice: ChatCompletionToolChoiceOptionParam | None = None,
        **api_llm_settings: Any,
    ) -> ChatCompletionAsyncStream[ChatCompletionChunk]:
        assert not api_tools, "Tool use is not supported in streaming mode"

        tools = api_tools or NOT_GIVEN
        tool_choice = api_tool_choice or NOT_GIVEN

        return await self._client.chat.completions.create(
            model=self._api_model_name,
            messages=api_messages,
            tools=tools,
            tool_choice=tool_choice,
            stream=True,
            **api_llm_settings,
        )



from collections.abc import Sequence
from copy import deepcopy
from typing import ClassVar, Generic, Protocol

from pydantic import BaseModel, TypeAdapter

from .generics_utils import AutoInstanceAttributesMixin
from .run_context import CtxT, RunContextWrapper, UserRunArgs
from .typing.content import ImageData
from .typing.io import (
    InT,
    LLMFormattedArgs,
    LLMFormattedSystemArgs,
    LLMPrompt,
    LLMPromptArgs,
)
from .typing.message import UserMessage


class DummySchema(BaseModel):
    pass


class FormatSystemArgsHandler(Protocol[CtxT]):
    def __call__(
        self,
        sys_args: LLMPromptArgs,
        *,
        ctx: RunContextWrapper[CtxT] | None,
    ) -> LLMFormattedSystemArgs: ...


class FormatInputArgsHandler(Protocol[InT, CtxT]):
    def __call__(
        self,
        *,
        usr_args: LLMPromptArgs,
        rcv_args: InT,
        batch_idx: int,
        ctx: RunContextWrapper[CtxT] | None,
    ) -> LLMFormattedArgs: ...


class PromptBuilder(AutoInstanceAttributesMixin, Generic[InT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {0: "_in_type"}

    def __init__(
        self,
        agent_id: str,
        sys_prompt: LLMPrompt | None,
        inp_prompt: LLMPrompt | None,
        sys_args_schema: type[LLMPromptArgs],
        usr_args_schema: type[LLMPromptArgs],
    ):
        self._in_type: type[InT]
        super().__init__()

        self._agent_id = agent_id
        self.sys_prompt = sys_prompt
        self.inp_prompt = inp_prompt
        self.sys_args_schema = sys_args_schema
        self.usr_args_schema = usr_args_schema
        self.format_sys_args_impl: FormatSystemArgsHandler[CtxT] | None = None
        self.format_inp_args_impl: FormatInputArgsHandler[InT, CtxT] | None = None

        self._rcv_args_type_adapter: TypeAdapter[InT] = TypeAdapter(self._in_type)

    def _format_sys_args(
        self,
        sys_args: LLMPromptArgs,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> LLMFormattedSystemArgs:
        if self.format_sys_args_impl:
            return self.format_sys_args_impl(sys_args=sys_args, ctx=ctx)

        return sys_args.model_dump(exclude_unset=True)

    def _format_inp_args(
        self,
        *,
        usr_args: LLMPromptArgs,
        rcv_args: InT,
        batch_idx: int = 0,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> LLMFormattedArgs:
        if self.format_inp_args_impl:
            return self.format_inp_args_impl(
                usr_args=usr_args, rcv_args=rcv_args, batch_idx=batch_idx, ctx=ctx
            )

        if not isinstance(rcv_args, BaseModel) and rcv_args is not None:
            raise TypeError(
                "Cannot apply default formatting to non-BaseModel received arguments."
            )

        usr_args_ = usr_args
        rcv_args_ = DummySchema() if rcv_args is None else rcv_args

        usr_args_dump = usr_args_.model_dump(exclude_unset=True)
        rcv_args_dump = rcv_args_.model_dump(exclude={"selected_recipient_ids"})

        return usr_args_dump | rcv_args_dump

    def make_sys_prompt(
        self,
        sys_args: LLMPromptArgs,
        *,
        ctx: RunContextWrapper[CtxT] | None,
    ) -> LLMPrompt | None:
        if self.sys_prompt is None:
            return None
        val_sys_args = self.sys_args_schema.model_validate(sys_args)
        fmt_sys_args = self._format_sys_args(val_sys_args, ctx=ctx)

        return self.sys_prompt.format(**fmt_sys_args)

    def _usr_messages_from_text(self, text: str) -> list[UserMessage]:
        return [UserMessage.from_text(text, model_id=self._agent_id)]

    def _usr_messages_from_content_parts(
        self, content_parts: Sequence[str | ImageData]
    ) -> Sequence[UserMessage]:
        return [UserMessage.from_content_parts(content_parts, model_id=self._agent_id)]

    def _usr_messages_from_rcv_args(
        self, rcv_args_batch: Sequence[InT]
    ) -> Sequence[UserMessage]:
        return [
            UserMessage.from_text(
                self._rcv_args_type_adapter.dump_json(
                    rcv,
                    exclude_unset=True,
                    indent=2,
                    exclude={"selected_recipient_ids"},
                    warnings="error",
                ).decode("utf-8"),
                model_id=self._agent_id,
            )
            for rcv in rcv_args_batch
        ]

    def _usr_messages_from_prompt_template(
        self,
        inp_prompt: LLMPrompt,
        usr_args: UserRunArgs | None = None,
        rcv_args_batch: Sequence[InT] | None = None,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> Sequence[UserMessage]:
        usr_args_batch_, rcv_args_batch_ = self._make_batched(usr_args, rcv_args_batch)

        val_usr_args_batch_ = [
            self.usr_args_schema.model_validate(u) for u in usr_args_batch_
        ]
        val_rcv_args_batch_ = [
            self._rcv_args_type_adapter.validate_python(rcv) for rcv in rcv_args_batch_
        ]

        formatted_inp_args_batch = [
            self._format_inp_args(
                usr_args=val_usr_args, rcv_args=val_rcv_args, batch_idx=i, ctx=ctx
            )
            for i, (val_usr_args, val_rcv_args) in enumerate(
                zip(val_usr_args_batch_, val_rcv_args_batch_, strict=False)
            )
        ]

        return [
            UserMessage.from_formatted_prompt(
                prompt_template=inp_prompt, prompt_args=inp_args
            )
            for inp_args in formatted_inp_args_batch
        ]

    def make_user_messages(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        usr_args: UserRunArgs | None = None,
        rcv_args_batch: Sequence[InT] | None = None,
        entry_point: bool = False,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> Sequence[UserMessage]:
        # 1) Direct user input (e.g. chat input)
        if chat_inputs is not None or entry_point:
            """
            * If chat inputs are provided, use them instead of the predefined
                input prompt template
            * In a multi-agent system, the predefined input prompt is used to
                construct agent inputs using the combination of received
                and user arguments.
                However, the first agent run (entry point) has no received
                messages, so we use the chat inputs directly, if provided.
            """
            if isinstance(chat_inputs, LLMPrompt):
                return self._usr_messages_from_text(chat_inputs)

            if isinstance(chat_inputs, Sequence) and chat_inputs:
                return self._usr_messages_from_content_parts(chat_inputs)

        # 2) No input prompt template + received args → raw JSON messages
        if self.inp_prompt is None and rcv_args_batch:
            return self._usr_messages_from_rcv_args(rcv_args_batch)

        # 3) Input prompt template + any args → batch & format
        if self.inp_prompt is not None:
            return self._usr_messages_from_prompt_template(
                inp_prompt=self.inp_prompt,
                usr_args=usr_args,
                rcv_args_batch=rcv_args_batch,
                ctx=ctx,
            )

        return []

    def _make_batched(
        self,
        usr_args: UserRunArgs | None = None,
        rcv_args_batch: Sequence[InT] | None = None,
    ) -> tuple[Sequence[LLMPromptArgs | DummySchema], Sequence[InT | DummySchema]]:
        usr_args_batch_ = (
            usr_args if isinstance(usr_args, list) else [usr_args or DummySchema()]
        )
        rcv_args_batch_ = rcv_args_batch or [DummySchema()]

        # Broadcast singleton → match lengths
        if len(usr_args_batch_) == 1 and len(rcv_args_batch_) > 1:
            usr_args_batch_ = [deepcopy(usr_args_batch_[0]) for _ in rcv_args_batch_]
        if len(rcv_args_batch_) == 1 and len(usr_args_batch_) > 1:
            rcv_args_batch_ = [deepcopy(rcv_args_batch_[0]) for _ in usr_args_batch_]
        if len(usr_args_batch_) != len(rcv_args_batch_):
            raise ValueError("User args and received args must have the same length")

        return usr_args_batch_, rcv_args_batch_



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Literal, TypeAlias, TypeVar

from pydantic import BaseModel, PrivateAttr, TypeAdapter

from ..generics_utils import AutoInstanceAttributesMixin

if TYPE_CHECKING:
    from ..run_context import CtxT, RunContextWrapper
else:
    CtxT = TypeVar("CtxT")

    class RunContextWrapper(Generic[CtxT]):
        """Runtime placeholder so RunContextWrapper[CtxT] works"""


_ToolInT = TypeVar("_ToolInT", bound=BaseModel, contravariant=True)  # noqa: PLC0105
_ToolOutT = TypeVar("_ToolOutT", covariant=True)  # noqa: PLC0105


class ToolCall(BaseModel):
    id: str
    tool_name: str
    tool_arguments: str


class BaseTool(
    AutoInstanceAttributesMixin, BaseModel, ABC, Generic[_ToolInT, _ToolOutT, CtxT]
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_schema",
        1: "_out_schema",
    }

    name: str
    description: str

    _in_schema: type[_ToolInT] = PrivateAttr()
    _out_schema: type[_ToolOutT] = PrivateAttr()

    # Supported by OpenAI API
    strict: bool | None = None

    @property
    def in_schema(self) -> type[_ToolInT]:  # type: ignore[reportInvalidTypeVarUse]
        # Exposing the type of a contravariant variable only, should be type safe
        return self._in_schema

    @property
    def out_schema(self) -> type[_ToolOutT]:
        return self._out_schema

    @abstractmethod
    async def run(
        self, inp: _ToolInT, ctx: RunContextWrapper[CtxT] | None = None
    ) -> _ToolOutT:
        pass

    async def __call__(
        self, ctx: RunContextWrapper[CtxT] | None = None, **kwargs: Any
    ) -> _ToolOutT:
        result = await self.run(self._in_schema(**kwargs), ctx=ctx)

        return TypeAdapter(self._out_schema).validate_python(result)


ToolChoice: TypeAlias = (
    Literal["none", "auto", "required"] | BaseTool[BaseModel, Any, Any]
)



from copy import deepcopy
from typing import Any, Literal, Optional, Protocol

from pydantic import ConfigDict, Field

from .memory import MessageHistory
from .run_context import RunContextWrapper
from .typing.io import AgentState, LLMPrompt

SetAgentStateStrategy = Literal["keep", "reset", "from_sender", "custom"]


class SetAgentState(Protocol):
    def __call__(
        self,
        cur_state: "LLMAgentState",
        *,
        rcv_state: AgentState | None,
        sys_prompt: LLMPrompt | None,
        ctx: RunContextWrapper[Any] | None,
    ) -> "LLMAgentState": ...


class LLMAgentState(AgentState):
    message_history: MessageHistory = Field(default_factory=MessageHistory)

    @property
    def batch_size(self) -> int:
        return self.message_history.batch_size

    @classmethod
    def from_cur_and_rcv_states(
        cls,
        cur_state: "LLMAgentState",
        *,
        rcv_state: Optional["AgentState"] = None,
        sys_prompt: LLMPrompt | None = None,
        strategy: SetAgentStateStrategy = "from_sender",
        set_agent_state_impl: SetAgentState | None = None,
        ctx: RunContextWrapper[Any] | None = None,
    ) -> "LLMAgentState":
        upd_mh = cur_state.message_history if cur_state else None
        if upd_mh is None or len(upd_mh) == 0:
            upd_mh = MessageHistory(sys_prompt=sys_prompt)

        if strategy == "keep":
            pass

        elif strategy == "reset":
            upd_mh.reset(sys_prompt)

        elif strategy == "from_sender":
            rcv_mh = (
                rcv_state.message_history
                if rcv_state and isinstance(rcv_state, "LLMAgentState")
                else None
            )
            if rcv_mh:
                upd_mh = deepcopy(rcv_mh)
            else:
                upd_mh.reset(sys_prompt)

        elif strategy == "custom":
            assert set_agent_state_impl is not None, (
                "Agent state setter implementation is not provided."
            )
            return set_agent_state_impl(
                cur_state=cur_state,
                rcv_state=rcv_state,
                sys_prompt=sys_prompt,
                ctx=ctx,
            )

        return cls.model_construct(message_history=upd_mh)

    def __repr__(self) -> str:
        return f"Message History: {len(self.message_history)}"

    model_config = ConfigDict(arbitrary_types_allowed=True)



# pyright: reportUnusedImport=false


from .agent_message import AgentMessage
from .base_agent import BaseAgent
from .comm_agent import CommunicatingAgent
from .llm import LLM, LLMSettings
from .llm_agent import LLMAgent
from .run_context import RunArgs, RunContextWrapper
from .typing.completion import Completion
from .typing.content import Content, ImageData
from .typing.io import AgentID, AgentState, LLMFormattedArgs, LLMPrompt, LLMPromptArgs
from .typing.message import AssistantMessage, Conversation, SystemMessage, UserMessage
from .typing.tool import BaseTool

__all__ = [
    "LLM",
    "AgentID",
    "AgentMessage",
    "AgentState",
    "AssistantMessage",
    "BaseAgent",
    "BaseTool",
    "CommunicatingAgent",
    "Completion",
    "Content",
    "Conversation",
    "ImageData",
    "LLMAgent",
    "LLMFormattedArgs",
    "LLMPrompt",
    "LLMPromptArgs",
    "LLMSettings",
    "RunArgs",
    "RunContextWrapper",
    "SystemMessage",
    "UserMessage",
]



import logging
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Sequence
from typing import Any, Generic, TypeVar, cast
from uuid import uuid4

from pydantic import BaseModel
from typing_extensions import TypedDict

from .memory import MessageHistory
from .typing.completion import Completion, CompletionChunk
from .typing.converters import Converters
from .typing.message import AssistantMessage, Conversation
from .typing.tool import BaseTool, ToolChoice

logger = logging.getLogger(__name__)


class LLMSettings(TypedDict):
    pass


SettingsT = TypeVar("SettingsT", bound=LLMSettings, covariant=True)  # noqa: PLC0105
ConvertT = TypeVar("ConvertT", bound=Converters, covariant=True)  # noqa: PLC0105


class LLM(ABC, Generic[SettingsT, ConvertT]):
    @abstractmethod
    def __init__(
        self,
        converters: ConvertT,
        model_name: str | None = None,
        model_id: str | None = None,
        llm_settings: SettingsT | None = None,
        tools: list[BaseTool[BaseModel, Any, Any]] | None = None,
        response_format: type | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__()

        self._converters = converters
        self._model_id = model_id or str(uuid4())[:8]
        self._model_name = model_name
        self._tools = {t.name: t for t in tools} if tools else None
        self._response_format = response_format
        self._llm_settings: SettingsT = llm_settings or cast("SettingsT", {})

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def model_name(self) -> str | None:
        return self._model_name

    @property
    def llm_settings(self) -> SettingsT:
        return self._llm_settings

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, Any]] | None:
        return self._tools

    @property
    def response_format(self) -> type | None:
        return self._response_format

    @tools.setter
    def tools(self, tools: list[BaseTool[BaseModel, Any, Any]] | None) -> None:
        self._tools = {t.name: t for t in tools} if tools else None

    @response_format.setter
    def response_format(self, response_format: type | None) -> None:
        self._response_format = response_format

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(model_id={self.model_id}; "
            f"model_name={self._model_name})"
        )

    @abstractmethod
    async def generate_completion(
        self,
        conversation: Conversation,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> Completion:
        pass

    @abstractmethod
    async def generate_completion_stream(
        self,
        conversation: Conversation,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[CompletionChunk]:
        pass

    @abstractmethod
    async def generate_message_batch(
        self,
        message_history: MessageHistory,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> Sequence[AssistantMessage]:
        pass
